#include <iostream>
#include <memory>
#include <sodium.h>

#include "point.hpp"

unsigned char ed25519_pk[crypto_sign_PUBLICKEYBYTES];
unsigned char ed25519_sk[crypto_sign_SECRETKEYBYTES];
unsigned char ed25519_seed[crypto_sign_SEEDBYTES];
unsigned char edwards25519_pk[crypto_scalarmult_ed25519_BYTES];
unsigned char edwards25519_sk[crypto_scalarmult_ed25519_SCALARBYTES];
unsigned char h[crypto_hash_sha512_BYTES];

template <typename T> struct Option {
  private:
    T value;
    bool is_some;

  public:
    Option() : is_some(false) {}
    Option(T value) : value(value), is_some(true) {}

    T unwrap() {
        if (!is_some)
            throw std::runtime_error("Option is empty");
        return value;
    }
};

// Templated alias for unique_ptr
template <typename T> using Box = std::unique_ptr<T>;

template <typename T, typename... Args> Box<T> Enbox(Args &&...args) {
    return std::make_unique<T>(std::forward<Args>(args)...);
}

int main(int argc, char *argv[]) {
    if (sodium_init() < 0) {
        return 1;
    }

    // create an Ed25519 keypair
    crypto_sign_keypair(ed25519_pk, ed25519_sk);

    // Extract the seed from the Ed25519 secret key
    crypto_sign_ed25519_sk_to_seed(ed25519_seed, ed25519_sk);

    Point p = Point(10, 5);
    std::cout << p << std::endl;
    p.move(5, 5);

    if (p.distance(Point(0, 0)) == p.distanceOrigin()) {
        std::cout << "Distance is the same" << std::endl;
    } else {
        std::cout << "Distance is not the same" << std::endl;
    }

    Box<Point> p_box = std::make_unique<Point>(10, 5);
    Box<Point> p_box2 = Enbox<Point>(10, 5);
    Box<Option<Point>> p_box3 = Enbox<Option<Point>>(Point(10, 5));

    p_box3->unwrap().move(5, 5);

    std::cout << p << std::endl;
    std::cout << *p_box << std::endl;
    std::cout << *p_box2 << std::endl;
    std::cout << "Hello, World!" << std::endl;
    return 0;
}
